obj/SaxHandlerBase.o: SaxHandlerBase.m SaxHandlerBase.h \
  /usr/include/GNUstep/Foundation/NSObject.h \
  /usr/include/GNUstep/Foundation/NSObjCRuntime.h \
  /usr/include/GNUstep/GNUstepBase/GSVersionMacros.h \
  /usr/include/GNUstep/GNUstepBase/preface.h \
  /usr/include/GNUstep/GNUstepBase/objc-gnu2next.h \
  /usr/include/GNUstep/GNUstepBase/GSConfig.h \
  /usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h \
  /usr/include/GNUstep/Foundation/NSZone.h \
  /usr/include/GNUstep/GNUstepBase/GNUstep.h \
  /usr/include/GNUstep/Foundation/NSDate.h ../SaxObjC/SaxEntityResolver.h \
  ../SaxObjC/SaxDTDHandler.h ../SaxObjC/SaxDocumentHandler.h \
  /usr/include/GNUstep/Foundation/NSString.h \
  /usr/include/GNUstep/Foundation/NSRange.h ../SaxObjC/SaxAttributeList.h \
  ../SaxObjC/SaxAttributes.h ../SaxObjC/SaxLocator.h \
  ../SaxObjC/SaxErrorHandler.h SaxException.h \
  /usr/include/GNUstep/Foundation/NSException.h

SaxHandlerBase.h:

/usr/include/GNUstep/Foundation/NSObject.h:

/usr/include/GNUstep/Foundation/NSObjCRuntime.h:

/usr/include/GNUstep/GNUstepBase/GSVersionMacros.h:

/usr/include/GNUstep/GNUstepBase/preface.h:

/usr/include/GNUstep/GNUstepBase/objc-gnu2next.h:

/usr/include/GNUstep/GNUstepBase/GSConfig.h:

/usr/include/GNUstep/GNUstepBase/GSObjCRuntime.h:

/usr/include/GNUstep/Foundation/NSZone.h:

/usr/include/GNUstep/GNUstepBase/GNUstep.h:

/usr/include/GNUstep/Foundation/NSDate.h:

../SaxObjC/SaxEntityResolver.h:

../SaxObjC/SaxDTDHandler.h:

../SaxObjC/SaxDocumentHandler.h:

/usr/include/GNUstep/Foundation/NSString.h:

/usr/include/GNUstep/Foundation/NSRange.h:

../SaxObjC/SaxAttributeList.h:

../SaxObjC/SaxAttributes.h:

../SaxObjC/SaxLocator.h:

../SaxObjC/SaxErrorHandler.h:

SaxException.h:

/usr/include/GNUstep/Foundation/NSException.h:
