(p
-jasdhf sdajfh asdfh asdfh ajasdhfj hasdjkfh jkasdf asdf jasd fhjasdhf jkasdhf
)p
(em
-asdklf
)em
(strong
-klasdf
)strong
(u
-asf
)u
(p
-asdkfj kl;asdfk asdjfjasdhf asdfkj daskjf asdkj flkasdf asdfkj   
)p
(ul
(li
(a
Ahref http://www.skyrix.de/
-SKYRiX
)a
)li
(li
(a
Ahref http://www.opengroupware.org/
-OGo
)a
)li
)ul
