/*
  Copyright (C) 2000-2005 SKYRIX Software AG

  This file is part of SOPE.

  SOPE is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  SOPE is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with SOPE; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#include "_WOStaticHTMLElement.h"
#include <NGObjWeb/WOResponse.h>
#include <NGObjWeb/WOContext.h>
#include "common.h"

@implementation _WOStaticHTMLElement

static Class StrClass = Nil;

+ (void)initialize {
  StrClass = [NSString class];
}

- (id)initWithBuffer:(const char *)_buffer length:(NSUInteger)_len {
  self->text = (_len > 0)
    ? [[StrClass alloc] initWithCString:_buffer length:_len]
    : nil;
  return self;
}
- (id)initWithCharacters:(const unichar *)_buffer length:(NSUInteger)_len {
  self->text = (_len > 0)
    ? [[StrClass alloc] initWithCharacters:_buffer length:_len]
    : nil;
  return self;
}

- (void)dealloc {
  [self->text release];
  [super dealloc];
}

/* generate response */

- (void)appendToResponse:(WOResponse *)_response inContext:(WOContext *)_ctx {
  if ([_ctx isRenderingDisabled] || self->text == nil) return;
  [_response appendContentString:self->text];
}

@end /* _WOStaticHTMLElement */
