/*
  Copyright (C) 2000-2005 SKYRIX Software AG

  This file is part of SOPE.

  SOPE is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  SOPE is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with SOPE; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#ifndef __WETableViewDefines_H__
#define __WETableViewDefines_H__

#define WETableView_              @"WETableView_"
#define WETableView_TitleMode     @"WETableView_TitleMode"
#define WETableView_ButtonMode    @"WETableView_ButtonMode"
#define WETableView_CollectMode   @"WETableView_CollectMode"
#define WETableView_HeaderMode    @"WETableView_HeaderMode"
#define WETableView_DataMode      @"WETableView_DataMode"
#define WETableView_FooterMode    @"WETableView_FooterMode"
#define WETableView_GroupMode     @"WETableView_GroupMode"

#define WETableView_INFOS         @"WETableView_INFOS"
#define WETableView_SORTEDKEY     @"WETableView_SORTEDKEY"
#define WETableView_ISDESCENDING  @"WETableView_ISDESCENDING"

#define WETableView_HasCollapseScript @"WETableView_HasCollapseScript"

// config stuff:
#define WETableView_titleColor    @"WETableView_titleColor"
#define WETableView_headerColor   @"WETableView_headerColor"
#define WETableView_footerColor   @"WETableView_footerColor"
#define WETableView_evenColor     @"WETableView_evenColor"
#define WETableView_oddColor      @"WETableView_oddColor"
#define WETableView_fontColor     @"WETableView_fontColor"
#define WETableView_fontFace      @"WETableView_fontFace"
#define WETableView_fontSize      @"WETableView_fontSize"

// sort icons:
#define WETableView_upwardIcon    @"WETableView_upwardIcon"
#define WETableView_downwardIcon  @"WETableView_downwardIcon"
#define WETableView_nonSortIcon   @"WETableView_nonSortIcon"

// navigation icons
#define WETableView_first           @"WETableView_first"
#define WETableView_first_blind     @"WETableView_first_blind"
#define WETableView_previous        @"WETableView_previous"
#define WETableView_previous_blind  @"WETableView_previous_blind"
#define WETableView_next            @"WETableView_next"
#define WETableView_next_blind      @"WETableView_next_blind"
#define WETableView_last            @"WETableView_last"
#define WETableView_last_blind      @"WETableView_last_blind"
#define WETableView_select_all      @"WETableView_select_all"
#define WETableView_deselect_all    @"WETableView_deselect_all"

// labels
#define WETableView_ofLabel         @"WETableView_ofLabel"
#define WETableView_toLabel         @"WETableView_toLabel"
#define WETableView_firstLabel      @"WETableView_firstLabel"
#define WETableView_previousLabel   @"WETableView_previousLabel"
#define WETableView_nextLabel       @"WETableView_nextLabel"
#define WETableView_lastLabel       @"WETableView_lastLabel"
#define WETableView_pageLabel       @"WETableView_pageLabel"
#define WETableView_sortLabel       @"WETableView_sortLabel"

#endif /* __WETableViewDefines_H__ */
